/* mkrawtape.c: make raw tape files from audio files
   Copyright (c) 2002 Darren Salt

   $Id$

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

   Author contact information:

   E-mail: linux@youmustbejoking.demon.co.uk

*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>


#define PROGNAME "mkrawtape"


static char *tmpsox = 0;	/* temporary filename */


/* Read a byte from a file */
static int
bget (FILE * in)
{
  int byte = fgetc (in);
  if (ferror (in))
  {
    fprintf (stderr, PROGNAME ": read failed: %s\n", strerror (errno));
    exit (2);
  }
  return byte;
}


/* Write a chunk to a file */
static void
mbget (FILE * in, void *buffer, size_t bytes)
{
  size_t got = fread (buffer, 1, bytes, in);
  if (got != bytes)
  {
    if (ferror (in))
      fprintf (stderr, PROGNAME ": read failed: %s\n", strerror (errno));
    else
      fprintf (stderr, PROGNAME ": unexpected EOF\n");
    exit (2);
  }
}


/* Write a byte to a file */
static void
bput (int byte, FILE * out)
{
  if (fputc (byte, out) != EOF)
    return;
  fprintf (stderr, PROGNAME ": write failed: %s\n", strerror (errno));
  exit (2);
}


/* Write a chunk to a file */
static void
mbput (FILE * out, const void *buffer, size_t bytes)
{
  if (fwrite (buffer, 1, bytes, out) == bytes)
    return;
  fprintf (stderr, PROGNAME ": write failed: %s\n", strerror (errno));
  exit (2);
}


/* Create a temporary filename */
static char *
get_tmp (void)
{
  char tmp[8];

  sprintf (tmp, "%04x", getpid ());
  return tempnam (0, tmp);	/* insecure, but the PID helps */
}


/* Delete the temporary file (atexit) */
static void
delete_tmp (void)
{
  if (tmpsox && unlink (tmpsox) && errno != ENOENT)
    fprintf (stderr, PROGNAME ": failed to delete %s: %s\n", tmpsox,
	     strerror (errno));
}


static void
usage (void)
{
  fprintf (stderr,
	   PROGNAME ": usage: " PROGNAME " [-r|-t] <VOC file> <RAW file>\n");
  exit (1);
}


static void
invoke_sox (char *arg1, ...)
{
  pid_t pid;
  char **argv = 0;
  int status, arga = 16;
  va_list ap;

  argv = calloc (arga, sizeof (char *));
  if (!argv)
  {
    fprintf (stderr, PROGNAME ": out of memory\n");
    exit (2);
  }
  argv[0] = "sox";
  argv[1] = arg1;

  if (arg1)
  {
    char *arg;
    int argc = 1;
    va_start (ap, arg1);
    do
    {
      arg = va_arg (ap, char *);
      if (++argc >= arga)
      {
	argv = realloc (argv, (arga *= 2) * sizeof (char *));
	if (!argv)
	{
	  fprintf (stderr, PROGNAME ": out of memory\n");
	  exit (2);
	}
      }
      argv[argc] = arg;
    }
    while (arg);
  }

  switch (pid = fork ())
  {
  case -1:
    fprintf (stderr, PROGNAME ": %s: couldn't run sox: %s\n", SOX, strerror (errno));
    exit (255);
  case 0:
    /* Child process */
    execv (SOX, argv);
    if (errno == ENOENT)
      fprintf (stderr, PROGNAME ": %s: sorry, no sox...\n", SOX);
    else
      fprintf (stderr, PROGNAME ": %s: couldn't run: %s\n", SOX, strerror (errno));
    exit (255);
  }

  atexit (delete_tmp);
  free (argv);
  do
  {
    wait (&status);
  }
  while (!WIFEXITED (status) && !WIFSIGNALED (status));
  if (WIFEXITED (status))
  {
    status = WEXITSTATUS (status);
    if (status)
    {
      if (status != 255)
	fprintf (stderr, PROGNAME ": sox exited with return code %i\n",
		 status);
      exit (2);
    }
  }
  else if (WIFSIGNALED (status))
  {
    fprintf (stderr, PROGNAME ": sox exited on signal %i\n",
	     WTERMSIG (status));
    exit (2);
  }
}


static FILE *
fopen_wb (const char *name)
{
  FILE *out = fopen (name, "wb");
  if (!out)
  {
    fprintf (stderr, PROGNAME ": %s: couldn't open for output: %s\n", name,
	     strerror (errno));
    return 0;
  }
  return out;
}


static int
write_raw (FILE * in, FILE ** outp, const char *outname)
{
  unsigned char buf[8] = { 0 };
  int bit = 0;
  size_t count = 0;
  size_t length = 0;
  FILE *out;

  *outp = out = fopen_wb (outname);
  if (!out)
    return 1;

  if (fwrite
      ("\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0Raw tape sample\0",
       64, 1, out) == EOF)
  {
    fprintf (stderr, PROGNAME ": header write failed: %s\n",
	     strerror (errno));
    return 1;
  }

#define WRITE(count) \
  { \
    if (count < 256) \
      bput (count, out); \
    else \
    { \
      buf[1] = count; \
      buf[2] = count >> 8; \
      buf[3] = count >> 16; \
      buf[4] = count >> 24; \
      mbput (out, buf, 5); \
    } \
  }

  fseek (in, 0, SEEK_END);
  length = ftell (in);
  fseek (in, 0, SEEK_SET);

  if (length)
  {
    bit = 1 - (bget (in) >> 7);
    count = 0;
  }
  while (length)
  {
    int nextbit = 0;
    do
    {
      count += (3500000 / 50000 / 7);
      if (--length)
	nextbit = bget (in) >> 7;
    }
    while (nextbit != bit && length);
    WRITE (count);
    count = 0;
    bit ^= 1;
  }
  if (count)
    WRITE (count);
  return 0;
}


static int
write_tzx (FILE * in, FILE ** outp, const char *outname)
{
  unsigned char buf[8];
  long tstates;
  size_t length;
  FILE *out;

  static const char header[] = {
    'Z', 'X', 'T', 'a', 'p', 'e', '!', 26, 0, 1,	/* header */
    0x15			/* first block ID */
  };

  mbget (in, buf, 8);
  if (memcmp (buf, "RIFF", 4))
    goto on_fire;

  mbget (in, buf, 8);
  if (memcmp (buf, "WAVEfmt ", 8))
    goto on_fire;

  mbget (in, buf, 8);
  if (memcmp (buf, "\x10\0\0\0\x01\0\x01\0", 6))
    goto on_fire;

  fseek (in, 0x18, SEEK_SET);
  mbget (in, buf, 8);
  tstates = buf[0] | buf[1] << 8 | buf[2] << 16 | buf[3] << 24;
  tstates = tstates ? 3500000 / tstates : -1;
  if (tstates < 1 || tstates > 32767)
  {
    fprintf (stderr, PROGNAME ": sample frequency is unrepresentable\n");
    return 1;
  }
  if (memcmp (buf, buf + 4, 4))
    goto on_fire;

  mbget (in, buf, 8);
  if (memcmp (buf, "\x01\0\x08\0data", 8))
    goto on_fire;

  mbget (in, buf, 4);
  length = buf[0] | buf[1] << 8 | buf[2] << 16 | buf[3] << 24;
  if ((length >> 3) >= 0x1000000)
  {
    fprintf (stderr, PROGNAME ": the sample is too big; complain loudly\n");
    return 1;
  }

  *outp = out = fopen_wb (outname);
  if (!out)
    return 1;

  mbput (out, header, sizeof (header));

  buf[0] = tstates;
  buf[1] = tstates >> 8;
  buf[2] = 1;
  buf[3] = 0;
  buf[4] = (length & 7) ? (length & 7) : 8;
  buf[5] = (length + 7) >> 3;
  buf[6] = (length + 7) >> 11;
  buf[7] = (length + 7) >> 19;
  mbput (out, buf, 8);

  while (length)
  {
    unsigned char byte;
    mbget (in, buf, (length >= 8) ? 8 : length);
    byte = (buf[0] & 128)
      | ((buf[1] & 128) >> 1) | ((buf[2] & 128) >> 2) | ((buf[3] & 128) >> 3)
      | ((buf[4] & 128) >> 4) | ((buf[5] & 128) >> 5) | ((buf[6] & 128) >> 6)
      | ((buf[7] & 128) >> 7);
    if (length >= 8)
    {
      length -= 8;
      bput (byte, out);
    }
    else
    {
      length = 0;
      bput (byte & (~255 >> length), out);
    }
  }
  return 0;

on_fire:
  fprintf (stderr, PROGNAME ": your sox are on fire\n");
  return 1;
}


int
main (int argc, char **argv)
{
  FILE *in, *out = 0;
  int ret;
  enum
  { FMT_RAW, FMT_TZX }
  format = FMT_RAW;

  if (argc != 3 && argc != 4)
    usage ();

  /* TODO: use getopt? */
  if (!strcmp (argv[1], "-r"))
  {
    format = FMT_RAW;
    argv++;
  }
  else if (!strcmp (argv[1], "-t"))
  {
    format = FMT_TZX;
    argv++;
  }
  else if (!strcmp (argv[1], "--"))
    argv++;
  else
    usage ();

  tmpsox = get_tmp ();
  if (!tmpsox)
  {
    fprintf (stderr, PROGNAME ": out of memory\n");
    return 2;
  }

  switch (format)
  {
  case FMT_RAW:
    invoke_sox (argv[1], "-traw", "-r50000", "-c1", "-u", "-p", tmpsox,
                "silence", "256", "256", "0%", "256", "256", "0%", 0);
    break;
  case FMT_TZX:
    /* force to mono ulin 8bit WAV (even if it already is) */
    invoke_sox (argv[1], "-twav", "-c1", "-u", "-p", tmpsox,
                "silence", "256", "256", "0%", "256", "256", "0%", 0);
    break;
  default:
    fprintf (stderr, PROGNAME ": eek, unimplemented output format!\n");
    return 2;
  }

  in = fopen (tmpsox, "rb");
  if (!in)
  {
    fprintf (stderr, PROGNAME ": %s: couldn't open for input: %s\n", tmpsox,
	     strerror (errno));
    return 2;
  }

  switch (format)
  {
  case FMT_RAW:
    ret = write_raw (in, &out, argv[2]);
    break;
  case FMT_TZX:
    ret = write_tzx (in, &out, argv[2]);
    break;
  default:
    fprintf (stderr, PROGNAME ": eek, unimplemented output format!\n");
    return 2;
  }

  if (fclose (in))
  {
    fprintf (stderr, PROGNAME ": %s: close failed: %s\n", tmpsox,
	     strerror (errno));
    return 2;
  }
  if (out && fclose (out))
  {
    fprintf (stderr, PROGNAME ": %s: close failed: %s\n", argv[2],
	     strerror (errno));
    return 2;
  }
  return ret ? 2 : 0;
}
